
require("imgui")

local masterPointsAmount
local masterPoints
local points
local maxpoints -- until the points are moved to the image data
local iterations
local speed -- the amount of iterations per frame

local data -- generated points
local image -- image of the generated points

local pause

local masterPointsColor
local pointsColor
local backgroundColor

local masterPointsSize
local drawMasterPointsNumbers
local clearOnMasterUpdate

local function randomPoint()
  return {
    x = math.random(1, love.graphics.getWidth() - 1),
    y = math.random(1, love.graphics.getHeight() - 1)
  }
end

local function midway(from, to)
  local dx = math.abs(from.x - to.x)
  local dy = math.abs(from.y - to.y)
  
  local minx = math.min(from.x, to.x)
  local miny = math.min(from.y, to.y)
  
  return {
    x = minx + (dx / 2),
    y = miny + (dy / 2)
  }
end

local function addPoint(algo)
  if #points == 0 then
    points[1] = randomPoint()
  end
  
  local point = masterPoints[math.random(1, #masterPoints)]
  local last = points[#points]
  
  local p = (algo or midway)(last, point)
  
  points[#points + 1] = p
  
  iterations = iterations + 1
end

local function clearCanvas()
  data = love.image.newImageData(love.graphics.getWidth(), love.graphics.getHeight())
  image = love.graphics.newImage(data)
  points = {}
  iterations = 0
end

local function generateMasterPoints(n)
  masterPoints = {}
  
  for i = 1, (n or masterPointsAmount) do
    masterPoints[i] = randomPoint()
  end
  
  if clearOnMasterUpdate then
    clearCanvas()
  end
end

local function getColor(c, invert)
  if invert then
    return (1 - c[1]) * 255, (1 - c[2]) * 255, (1 - c[3]) * 255
  end
  
  return c[1] * 255, c[2] * 255, c[3] * 255
end

function love.load()
  math.randomseed(os.time())
  
  iterations = 0
  speed = 1000
  masterPointsAmount = 3
  masterPointsColor = { 0, 0, 1 }
  pointsColor = { 0, 1, 0 }
  masterPointsSize = 10
  backgroundColor = { 0.3, 0.3, 0.3 }
  drawMasterPointsNumbers = false
  clearOnMasterUpdate = true
  maxpoints = 100
  pause = true
  points = {randomPoint()}
  
  clearCanvas()
  generateMasterPoints()
end

function love.update(dt)
  imgui.NewFrame()
  
  if pause then return end
  
  if #points < maxpoints then -- generate new points
    for i = 1, speed do
      addPoint()
    end
  else -- write the buffer to the image data
    local last = points[#points]
    
    for i = #points, 1, -1 do
      data:setPixel(points[i].x, points[i].y, getColor(pointsColor))
      points[i] = nil
    end
    
    points[1] = last -- needed for the algorithm
    image = love.graphics.newImage(data) -- update the image
  end
end

function love.draw()
  love.graphics.setBackgroundColor(getColor(backgroundColor))
  love.graphics.setColor({ 255, 255, 255 })
  
  love.graphics.draw(image, 0, 0) -- generated data
  
  love.graphics.setPointSize(1)
  love.graphics.setColor(getColor(pointsColor))
  
  for _, p in ipairs(points) do -- last {maxpoints} generated points
    love.graphics.points(p.x, p.y)
  end
  
  love.graphics.setPointSize(masterPointsSize)
  love.graphics.setColor(getColor(masterPointsColor))
  
  for i, p in ipairs(masterPoints) do -- the master set of points
    love.graphics.points(p.x, p.y)
    
    if drawMasterPointsNumbers then
      love.graphics.setColor(getColor(masterPointsColor, true)) -- to better see the numbers
      local offset = masterPointsSize / 2
      love.graphics.rectangle("fill", p.x + offset, p.y + offset, love.graphics.getFont():getWidth(tostring(i)), love.graphics.getFont():getHeight())
      love.graphics.setColor(getColor(masterPointsColor))
      love.graphics.print(tostring(i), p.x + offset, p.y + offset)
    end
  end
  
  do -- imgui
    love.graphics.setColor({ 255, 255, 255 })
    
    local status

    imgui.Text("FPS: "..love.timer.getFPS())
    imgui.Text("Iterations: "..tostring(iterations))
    
    status, backgroundColor[1], backgroundColor[2], backgroundColor[3] = imgui.ColorEdit3("Background Color", backgroundColor[1], backgroundColor[2], backgroundColor[3])
    status, pointsColor[1], pointsColor[2], pointsColor[3] = imgui.ColorEdit3("Points Color", pointsColor[1], pointsColor[2], pointsColor[3])
    status, masterPointsColor[1], masterPointsColor[2], masterPointsColor[3] = imgui.ColorEdit3("Master Points Color", masterPointsColor[1], masterPointsColor[2], masterPointsColor[3])
    
    status, masterPointsAmount = imgui.SliderInt("Master Points", masterPointsAmount, 1, 20)
    status, masterPointsSize = imgui.SliderInt("Master Points Size", masterPointsSize, 1, 20)
    
    status, speed = imgui.SliderInt("Iterations per frame", speed, 1, 50000)
    
    status, drawMasterPointsNumbers = imgui.Checkbox("Draw Master Points Numbers", drawMasterPointsNumbers)
    
    status, clearOnMasterUpdate = imgui.Checkbox("Clear on Generate", clearOnMasterUpdate)
    
    status, pause = imgui.Checkbox("Active", not pause)
    pause = not pause
    
    imgui.Text("Press keys 1 to 9 to move the master points to the mouse")
    
    if imgui.Button("Clear Canvas") then
      iterations = 0
      clearCanvas()
    end
    
    if imgui.Button("Generate Master Points") then
      generateMasterPoints()
    end

    imgui.Render()
  end
end

function love.textinput(t)
    imgui.TextInput(t)
    if not imgui.GetWantCaptureKeyboard() then
        -- Pass event to the game
    end
end

local function moveMaster(key, value) -- move a master point to the mouse
  if key == value then
    local n = tonumber(value)
    if #masterPoints >= n then
      masterPoints[n].x = love.mouse.getX()
      masterPoints[n].y = love.mouse.getY()
    end
    if clearOnMasterUpdate then
      clearCanvas()
    end
  end
end

function love.keypressed(key)
  imgui.KeyPressed(key)
  if not imgui.GetWantCaptureKeyboard() then
    -- Pass event to the game
    if key == "return" then addPoint() end

    if key == "p" then pause = not pause end
    
    moveMaster(key, "1")
    moveMaster(key, "2")
    moveMaster(key, "3")
    moveMaster(key, "4")
    moveMaster(key, "5")
    moveMaster(key, "6")
    moveMaster(key, "7")
    moveMaster(key, "8")
    moveMaster(key, "9")
  end
end

function love.keyreleased(key)
  imgui.KeyReleased(key)
  if not imgui.GetWantCaptureKeyboard() then
    -- Pass event to the game
  end
end

function love.mousemoved(x, y)
  imgui.MouseMoved(x, y)
  if not imgui.GetWantCaptureMouse() then
    -- Pass event to the game
  end
end

function love.mousepressed(x, y, button)
  imgui.MousePressed(button)
  if not imgui.GetWantCaptureMouse() then
    -- Pass event to the game
  end
end

function love.mousereleased(x, y, button)
  imgui.MouseReleased(button)
  if not imgui.GetWantCaptureMouse() then
    -- Pass event to the game
  end
end

function love.wheelmoved(x, y)
  imgui.WheelMoved(y)
  if not imgui.GetWantCaptureMouse() then
    -- Pass event to the game
  end
end
